	
	----------------------------------------------------
	-- Mezzonic Cache Helper 1.04 by Fail US-Ragnaros --
	----------------------------------------------------

local ADDON_NAME,MezzonicCacheHelper = ...

--Create the Main Frame
MezzonicCacheHelper.Frame = MezzonicCacheHelper:FrameTemplate({
	hide = true,
	name = "MezzonicCacheHelperFrame",
	width = 350,
	height = 480,
	backdrop = {
		color = {0,0,0,.75},
		border = {1,1,1,1},
		edgeFile = "Interface\\Tooltips\\UI-Tooltip-Border",
		edgeSize = 16,
		insets = {left=4,right=4,top=4,bottom=4},
	},
	text = {
		{
			name = "Description",
			text = string.format("|cffffd000%s v|r%s\n|cffffd000#|r1 |cffffd000Copy|r the |cffffd000Puzzle|r on the |cffffd000Grid|r above.\n|cffffd000#|r2 Click |cffffd000Solve|r\n|cffffd000#|r3 Click the |cffffd000Puzzle|r as shown in |cff00c000green|r.\n\n|cffffd000Created by|r %s",GetAddOnMetadata(ADDON_NAME,"Title"),GetAddOnMetadata(ADDON_NAME,"Version"),GetAddOnMetadata(ADDON_NAME,"Author")),
			point = "BOTTOMLEFT",
			xOff = 8,
			yOff = 8,
			size = 14,
			flag = "OUTLINE",
			horizontal = "LEFT",
			vertical = "TOP",
		},
		{
			name = "Percent",
			text = "%",
			point = "BOTTOMRIGHT",
			xOff = -8,
			yOff = 8,
			size = 14,
			flag = "OUTLINE",
			horizontal = "RIGHT",
			vertical = "TOP",
		},
	},
	movable = true,
	screen = true,
	script = {
		{
			type = "OnShow",
			func = function(self)
				self.ButtonSolve:Enable()
				self.ButtonSolve.Text:SetTextColor(1,1,1,1)
				self.ButtonSolve:SetBackdropBorderColor(0,0,0,1)
				for row=1,5 do
					for col=1,5 do
						self.Input[row][col] = 0
					end
				end
				self:Update(self.Input,true)
				self:ClearAllPoints()
				self:SetPoint(MezzonicCacheHelperDB.point,UIParent,MezzonicCacheHelperDB.relativePoint,MezzonicCacheHelperDB.xOff,MezzonicCacheHelperDB.yOff)
				self:SetScale(MezzonicCacheHelperDB.scale)
				self.Scale:SetText(MezzonicCacheHelperDB.scale*100)
			end,
		},
		{
			type = "OnMouseUp",
			func = function(self,button)
				if button ~= "LeftButton" then return end
				local point,_,relativePoint,xOff,yOff = self:GetPoint()
				MezzonicCacheHelperDB.point = point
				MezzonicCacheHelperDB.relativePoint = relativePoint
				MezzonicCacheHelperDB.xOff = xOff
				MezzonicCacheHelperDB.yOff = yOff
			end,
		},
	},
	event = {
		{
			event = "ADDON_LOADED",
			func = function(self,event,addonName)
				if addonName ~= ADDON_NAME then return end
				self:UnregisterEvent(event)
				local defaultDB = {scale = 1,point = "RIGHT",relativePoint = "RIGHT",xOff = -371,yOff = 40}
				if MezzonicCacheHelperDB == nil then
					MezzonicCacheHelperDB = defaultDB
				else
					for key,value in pairs(defaultDB) do
						if MezzonicCacheHelperDB[key] == nil then
							MezzonicCacheHelperDB[key] = value
						end
					end
				end
			end,
		},
		{
			event = {"PLAYER_ENTERING_WORLD","ZONE_CHANGED"},
			func = function(self)
				local mapID = C_Map.GetBestMapForUnit("player")
				--check if we are on Zereth Mortis
				if mapID == 1970 then
					self:RegisterUnitEvent("UNIT_AURA","player")
					self:Lock()
				else --if not unregister UNIT_AURA
					self:UnregisterEvent("UNIT_AURA")
				end
			end,
		},
		{
			event = "UNIT_AURA",
			unregister = true,
			func = function(self,_,unitID)
				if unitID ~= "player" then return end
				self:Lock()
			end,
		},
	},
})
MezzonicCacheHelper.Frame.Output = {{0,0,0,0,0},{0,0,0,0,0},{0,0,0,0,0},{0,0,0,0,0},{0,0,0,0,0}}
MezzonicCacheHelper.Frame.Input = {{0,0,0,0,0},{0,0,0,0,0},{0,0,0,0,0},{0,0,0,0,0},{0,0,0,0,0}}
MezzonicCacheHelper.Frame.Pattern = {{{1,0,1,0,1},{1,0,1,0,1},{0,0,0,0,0},{1,0,1,0,1},{1,0,1,0,1}},{{1,1,0,1,1},{0,0,0,0,0},{1,1,0,1,1},{0,0,0,0,0},{1,1,0,1,1}},{{0,1,1,1,0},{1,0,1,0,1},{1,1,0,1,1},{1,0,1,0,1},{0,1,1,1,0}},{{},{},{},{},{}}}
MezzonicCacheHelper.Frame.Check = {{},{},{},{},{}}
MezzonicCacheHelper.Frame.Grid = {{},{},{},{},{}}
MezzonicCacheHelper.Frame.Moves = 0
MezzonicCacheHelper.Frame.Quiet = 0

--Simple Scale
MezzonicCacheHelper.Frame.Scale = MezzonicCacheHelper:FrameTemplate({
	type = "EDITBOX",
	parent = MezzonicCacheHelper.Frame,
	point = "BOTTOMRIGHT",
	xOff = -15,
	yOff = -9,
	width = 64,
	height = 48,
	numeric = true,
	horizontal = "RIGHT",
	insets = {8,8,8,8},
	script = {
		{
			type = "OnEscapePressed",
			func = function(self)
				self:ClearFocus()
			end,
		},
		{
			type = "OnEnterPressed",
			func = function(self)
				self:ClearFocus()
			end,
		},
		{
			type = "OnEditFocusLost",
			func = function(self)
				local value = math.floor(math.min(math.max(self:GetNumber(),75),150))
				self:SetText(value)
				self:GetParent():SetScale(value/100)
				MezzonicCacheHelperDB.scale = value/100
			end,
		},
	},
})
MezzonicCacheHelper.Frame.Percent:ClearAllPoints()
MezzonicCacheHelper.Frame.Percent:SetPoint("LEFT",MezzonicCacheHelper.Frame.Scale,"RIGHT",-8,0)

--Create the Grid
for index=0,24 do
	MezzonicCacheHelper.Frame["ButtonGrid"..index+1] = MezzonicCacheHelper:FrameTemplate({
		type = "BUTTON",
		parent = MezzonicCacheHelper.Frame,
		point = "TOPLEFT",
		xOff = ((index%5)*64)+16,
		yOff = (math.floor(index/5)*-64)-16,
		width = 64,
		height = 64,
		backdrop = {
			color = {0,0,0,0},
			border = {1,1,1,1},
		},
		script = {
			{
				type = "OnClick",
				func = function(self)
					local parent = self:GetParent()
					parent.Input[self.row][self.col] = (parent.Input[self.row][self.col]+1)%2
					if parent.Input[self.row][self.col] == 1 then
						self:SetBackdropColor(1,1,1,1)
						self:SetBackdropBorderColor(0,0,0,1)
					else
						self:SetBackdropColor(0,0,0,1)
						self:SetBackdropBorderColor(1,1,1,1)
					end
				end,
			},
		},
	})
	MezzonicCacheHelper.Frame["ButtonGrid"..index+1].row = math.floor(index/5)+1
	MezzonicCacheHelper.Frame["ButtonGrid"..index+1].col = (index%5)+1
	MezzonicCacheHelper.Frame.Grid[math.floor(index/5)+1][(index%5)+1] = MezzonicCacheHelper.Frame["ButtonGrid"..index+1]
end

--Solve Button
MezzonicCacheHelper.Frame.ButtonSolve = MezzonicCacheHelper:FrameTemplate({
	type = "BUTTON",
	parent = MezzonicCacheHelper.Frame,
	point = "TOPLEFT",
	relative = MezzonicCacheHelper.Frame.ButtonGrid21,
	relativePoint = "BOTTOMLEFT",
	yOff = -8,
	width = 152,
	height = 24,
	backdrop = {
		color = {.1,.1,.1,1},
		border = {0,0,0,1},
	},
	text = {
		{
			text = "Solve",
			size = 16,
			flag = "OUTLINE",
		},
	},
	script = {
		{
			type = "OnEnter",
			func = function(self)
				if self.Unsolvable then return end
				self:SetBackdropBorderColor(0,.75,0,1)
				self.Text:SetTextColor(0,.75,0,1)
			end,
		},
		{
			type = "OnLeave",
			func = function(self)
				self:SetBackdropBorderColor(0,0,0,1)
				self.Text:SetTextColor(1,1,1,1)
				if self.Unsolvable then
					local parent = self:GetParent()
					parent.Update(parent,parent.Input,true)
					self.Unsolvable = nil
				end
			end,
		},
		{
			type = "OnClick",
			func = function(self)
				PlaySound(687,"Master")
				local parent = self:GetParent()
				self:Disable()
				self:SetBackdropBorderColor(0,.75,0,1)
				self.Text:SetTextColor(0,.75,0,1)
				self.Moves = 99
				--every Mezzonic puzzle in-game can be solved on 5 moves
				--but Chasing the Lights can give us a ton of extra moves
				--so we just spin and mirror the puzzle
				--solving it in different directions can give us less moves
				for index=1,8 do
					for row=1,5 do
						for col=1,5 do
							if index == 1 then --top -> bottom
								self.Input[row][col] = parent.Input[row][col]+0
							elseif index == 2 then --top -> bottom (mirror)
								self.Input[row][6-col] = parent.Input[row][col]+0
							elseif index == 3 then --bottom -> top
								self.Input[6-row][col] = parent.Input[row][col]+0
							elseif index == 4 then --bottom -> top (mirror)
								self.Input[6-row][6-col] = parent.Input[row][col]+0
							elseif index == 5 then --left -> right
								self.Input[col][6-row] = parent.Input[row][col]+0
							elseif index == 6 then --left -> right (mirror)
								self.Input[col][row] = parent.Input[row][col]+0
							elseif index == 7 then --right -> left
								self.Input[6-col][6-row] = parent.Input[row][col]+0
							elseif index == 8 then --right -> left (mirror)
								self.Input[6-col][row] = parent.Input[row][col]+0
							end
						end
					end
					parent:Solve(self.Input)
					--if it has less moves that the previous ones then rwe want it
					if parent.Moves < self.Moves then
						self.Moves = parent.Moves
						--we then need to spin and mirror the output
						for row=1,5 do
							for col=1,5 do
								if index == 1 then
									self.Output[row][col] = parent.Output[row][col]+0
								elseif index == 2 then
									self.Output[row][col] = parent.Output[row][6-col]+0
								elseif index == 3 then
									self.Output[row][col] = parent.Output[6-row][col]+0
								elseif index == 4 then
									self.Output[row][col] = parent.Output[6-row][6-col]+0
								elseif index == 5 then
									self.Output[row][col] = parent.Output[col][6-row]+0
								elseif index == 6 then
									self.Output[row][col] = parent.Output[col][row]+0
								elseif index == 7 then
									self.Output[row][col] = parent.Output[6-col][6-row]+0
								elseif index == 8 then
									self.Output[row][col] = parent.Output[6-col][row]+0
								end
							end
						end
					end
				end
				--if 99 is the moves required that means this puzzle has no solution
				--we will just turn the grid red on the next function
				parent:Update(self.Moves == 99 and parent.Input or self.Output,false,self.Moves == 99)
			end,
		},
	},
})
MezzonicCacheHelper.Frame.ButtonSolve.Input = {{},{},{},{},{}}
MezzonicCacheHelper.Frame.ButtonSolve.Output = {{0,0,0,0,0},{0,0,0,0,0},{0,0,0,0,0},{0,0,0,0,0},{0,0,0,0,0}}
MezzonicCacheHelper.Frame.ButtonSolve.Moves = 99

--Reset Button
MezzonicCacheHelper.Frame.ButtonReset = MezzonicCacheHelper:FrameTemplate({
	type = "BUTTON",
	parent = MezzonicCacheHelper.Frame,
	point = "TOPLEFT",
	relative = MezzonicCacheHelper.Frame.ButtonSolve,
	relativePoint = "TOPRIGHT",
	xOff = 16,
	width = 152,
	height = 24,
	backdrop = {
		color = {.1,.1,.1,1},
		border = {0,0,0,1},
	},
	text = {
		{
			text = "Reset",
			size = 16,
			flag = "OUTLINE",
		},
	},
	script = {
		{
			type = "OnEnter",
			func = function(self)
				self:SetBackdropBorderColor(.75,0,0,1)
				self.Text:SetTextColor(.75,0,0,1)
			end,
		},
		{
			type = "OnLeave",
			func = function(self)
				self:SetBackdropBorderColor(0,0,0,1)
				self.Text:SetTextColor(1,1,1,1)
			end,
		},
		{
			type = "OnClick",
			func = function(self)
				PlaySound(687,"Master")
				local parent = self:GetParent()
				parent.ButtonSolve:Enable()
				parent.ButtonSolve:SetBackdropBorderColor(0,0,0,1)
				parent.ButtonSolve.Text:SetTextColor(1,1,1,1)
				for row=1,5 do
					for col=1,5 do
						parent.Input[row][col] = 0
					end
				end
				parent:Update(parent.Input,true)
			end,
		},
	},
})

function MezzonicCacheHelper.Frame:Lock()
	--check if we are on a Mezzonic Protolock
	for index=1,255 do
		local _,_,_,_,_,_,_,_,_,spellID = UnitAura("player",index)
		if not spellID then break end
		if spellID == 351405 or spellID == 366042 or spellID == 366106 then
			self:Show()
			return
		end
	end
	self:Hide()
end

function MezzonicCacheHelper.Frame:Toggle(grid,row,col)
	--we toggle the lights from the row/col and adjacent cells
	if grid[row-1] and grid[row-1][col] then grid[row-1][col] = (grid[row-1][col]+1)%2 end
    if grid[row] and grid[row][col+1] then grid[row][col+1] = (grid[row][col+1]+1)%2 end
    if grid[row] and grid[row][col-1] then grid[row][col-1] = (grid[row][col-1]+1)%2 end
    if grid[row+1] and grid[row+1][col] then grid[row+1][col] = (grid[row+1][col]+1)%2 end
    if grid[row] and grid[row][col] then grid[row][col] = (grid[row][col]+1)%2 end
end

--Chasing the Light method, basicaly brute-forcing with a sure method
function MezzonicCacheHelper.Frame:Solve(input)
	--we need to simulate solving the grid using the Chase The Lights method
	self.Moves = 0 --number of moves required to complete the puzzle
	for row=1,5 do
		for col=1,5 do
			self.Check[row][col] = input[row][col]+0 --we need to do a double-check after we are over
			self.Output[row][col] = 0 --which moves we need to make
		end
	end
	for iteration=1,2 do --we go across the grid twice
		for row=1,5 do
			for col=1,5 do
				if input[row][col] == 1 then
					if row == 5 then
						--row 5 has special rules
						if iteration == 2 or col >= 4 then break end
						--depending on which column is ON we toggle on row 1
						if col == 1 then
							self:Toggle(input,1,4)
							self:Toggle(input,1,5)
						elseif col == 2 then
							self:Toggle(input,1,2)
							self:Toggle(input,1,5)
						elseif col == 3 then
							self:Toggle(input,1,4)
						end
					else
						--if a light is ON then we do two things
						--we need to toggle the light on the next row directly below that one
						if input[row+1] and input[row+1][col] then self:Toggle(input,row+1,col) end
						--we also save this move to the output
						if self.Output[row+1] and self.Output[row+1][col] then self.Output[row+1][col] = (self.Output[row+1][col]+1)%2 end
					end
				end
			end
		end
	end
	--we then use our output to toggle the lights on the double-check grid
	for row=1,5 do
		for col=1,5 do
			if self.Output[row][col] == 1 then
				self.Moves = self.Moves+1
				self:Toggle(self.Check,row,col)
			end
		end
	end
	--then we check row 2 of the double-check grid for any ON lights
	for col=1,5 do
		if self.Check[2][col] == 1 then --we find a ON light
			self.Moves = self.Moves+1
			self:Toggle(self.Check,1,col) --we toggle the light above it
			self.Output[1][col] = 1 --we add it to the output
		end
	end
	--to finish we use quiet patterns that doesn't change the state of the grid to get rid of unnecessary moves
	for pattern=1,3 do
		self.Quiet = 0
		for row=1,5 do
			for col=1,5 do
				self.Pattern[4][row][col] = (self.Output[row][col]+self.Pattern[pattern][row][col])%2
				self.Quiet = self.Quiet+self.Pattern[4][row][col]
			end
		end
		--if the quiet output is better then we want it
		if self.Quiet < self.Moves then
			self.Moves = self.Quiet
			for row=1,5 do
				for col=1,5 do
					self.Output[row][col] = self.Pattern[4][row][col]+0
				end
			end
		end
	end
	--after all of this we check if there are still any ON light
	for row=1,5 do
		for col=1,5 do
			if self.Check[row][col] == 1 then --if we find any ON light it means the grid has no solution
				self.Moves = 99
				break
			end
		end
	end
end

function MezzonicCacheHelper.Frame:Update(grid,enable,unsolvable)
	if unsolvable then --if the grid cannot be solved then we turn everything red
		self.ButtonSolve:Enable()
		self.ButtonSolve:SetBackdropBorderColor(.75,0,0,1)
		self.ButtonSolve.Text:SetTextColor(.75,0,0,1)
		self.ButtonSolve.Unsolvable = true
	end
	--update the grid according to the input
	for row=1,5 do
		for col=1,5 do
			self.Grid[row][col]:EnableMouse(enable)
			if grid[row][col] == 1 then
				if enable then --if the grid is enabled then ON squares should be white
					self.Grid[row][col]:SetBackdropColor(1,1,1,1)
					self.Grid[row][col]:SetBackdropBorderColor(0,0,0,1)
				else
					if unsolvable then --if the grid cannot be solved then we turn everything red
						self.Grid[row][col]:SetBackdropColor(.75,0,0,1)
					else --else the grid is green
						self.Grid[row][col]:SetBackdropColor(0,.75,0,1)
					end
					self.Grid[row][col]:SetBackdropBorderColor(1,1,1,1)
				end
			else
				self.Grid[row][col]:SetBackdropColor(0,0,0,1)
				self.Grid[row][col]:SetBackdropBorderColor(1,1,1,1)
			end
		end
	end
end